package net.jfwu.modeling.utils;

import static com.google.common.collect.ImmutableList.builder;
import static java.util.Arrays.asList;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Collection;

import net.jfwu.modeling.Child;
import net.jfwu.modeling.Children;
import net.jfwu.modeling.Parent;

final class ModelDirectExtractionUtils {
	private ModelDirectExtractionUtils() {
	}

	public static <S, T> T getChild(S model, Class<S> modelClass,
			Class<T> childClass) {
		return get("child", model, modelClass, childClass,
				getChildField(modelClass, childClass),
				getChildMethod(modelClass, childClass));
	}

	@SuppressWarnings("unchecked")
	public static <S, T> Collection<T> getChildren(S model,
			Class<S> modelClass, Class<T> childrenClass) {
		return get("children", model, modelClass, Collection.class,
				getChildrenField(modelClass, childrenClass),
				getChildrenMethod(modelClass, childrenClass));
	}

	public static <S, T> T getParent(S model, Class<S> modelClass,
			Class<T> parentClass) {
		return get("child", model, modelClass, parentClass,
				getParentField(modelClass, parentClass),
				getParentMethod(modelClass, parentClass));
	}

	private static <S, T> T get(String type, S model, Class<S> modelClass,
			Class<T> obtainedClass, Field field, Method method) {
		try {
			if (field != null) {
				field.setAccessible(true);
				return (T) field.get(model);
			}
			if (method != null) {
				method.setAccessible(true);
				return (T) method.invoke(model);
			}
		} catch (IllegalAccessException | IllegalArgumentException
				| InvocationTargetException e) {
			throw new IllegalStateException(e);
		}
		throw new IllegalArgumentException("Model "
				+ modelClass.getSimpleName() + "does not declare " + type
				+ " of class " + obtainedClass.getSimpleName());
	}

	private static Field getChildField(Class<?> modelClass, Class<?> childClass) {
		for (Field field : merge(modelClass.getFields(),
				modelClass.getDeclaredFields())) {
			Child annotation = field.getAnnotation(Child.class);
			if (isRightField(field, childClass, annotation)
					&& annotation.value() == childClass)
				return field;
		}
		return null;
	}

	private static Method getChildMethod(Class<?> modelClass,
			Class<?> childClass) {
		for (Method method : merge(modelClass.getMethods(),
				modelClass.getDeclaredMethods())) {
			Child annotation = method.getAnnotation(Child.class);
			if (isRightGetter(method, childClass, annotation)
					&& annotation.value() == childClass)
				return method;
		}
		return null;
	}

	private static Field getChildrenField(Class<?> modelClass,
			Class<?> childrenClass) {
		for (Field field : merge(modelClass.getFields(),
				modelClass.getDeclaredFields())) {
			Children annotation = field.getAnnotation(Children.class);
			if (isRightField(field, childrenClass, annotation)
					&& annotation.type() == childrenClass)
				return field;
		}
		return null;
	}

	private static Method getChildrenMethod(Class<?> modelClass,
			Class<?> childrenClass) {
		for (Method method : merge(modelClass.getMethods(),
				modelClass.getDeclaredMethods())) {
			Children annotation = method.getAnnotation(Children.class);
			if (isRightGetter(method, childrenClass, annotation)
					&& annotation.type() == childrenClass)
				return method;
		}
		return null;
	}

	private static Field getParentField(Class<?> modelClass,
			Class<?> parentClass) {
		for (Field field : merge(modelClass.getFields(),
				modelClass.getDeclaredFields())) {
			Parent annotation = field.getAnnotation(Parent.class);
			if (isRightField(field, parentClass, annotation)
					&& annotation.value() == parentClass)
				return field;
		}
		return null;
	}

	private static Method getParentMethod(Class<?> modelClass,
			Class<?> parentClass) {
		for (Method method : merge(modelClass.getMethods(),
				modelClass.getDeclaredMethods())) {
			Parent annotation = method.getAnnotation(Parent.class);
			if (isRightGetter(method, parentClass, annotation)
					&& annotation.value() == parentClass)
				return method;
		}
		return null;
	}

	private static boolean isRightField(Field field, Class<?> clazz,
			Object annotation) {
		return field.getType().isAssignableFrom(clazz) && annotation != null;
	}

	private static boolean isRightGetter(Method method, Class<?> clazz,
			Object annotation) {
		return method.getParameterTypes().length == 0
				&& method.getReturnType().isAssignableFrom(clazz)
				&& annotation != null;
	}

	private static <T> T[] merge(T[] a1, T[] a2) {
		return builder().addAll(asList(a1)).addAll(asList(a2)).build()
				.toArray(a1);
	}
}
