/**
 * Copyright 2013 Mateusz Kubuszok
 *
 * <p>Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at</p> 
 * 
 * <p>http://www.apache.org/licenses/LICENSE-2.0</p>
 *
 * <p>Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.</p>
 */
package net.jfwu.synchronization;

import static java.lang.Integer.toHexString;
import static java.lang.System.identityHashCode;

import java.lang.reflect.Method;

import javassist.util.proxy.MethodHandler;

/**
 * Common logic for SynchronizingProxyHandler and PassingProxyHandler.
 * 
 * @see net.jfwu.synchronization.PassingProxyHandler
 * @see net.jfwu.synchronization.SynchronizingProxyHandler
 */
abstract class AbstractProxyHandler implements MethodHandler {
	private final SynchronizingProxyController proxyDB;

	private final Object object;

	AbstractProxyHandler(SynchronizingProxyController proxyDB, Object object) {
		this.proxyDB = proxyDB;
		this.object = object;
	}

	/**
	 * Executes call and wraps result in proxy if possible.
	 * 
	 * @param proxy
	 *            handled proxy object
	 * @param method
	 *            method to invoke
	 * @param args
	 *            argument to pass
	 * @return result, potentially proxy
	 * @throws Throwable
	 *             thrown if invocation failed
	 */
	protected Object executeCall(Object proxy, Method method, Object[] args)
			throws Throwable {
		Object value = evaluate(proxy, method, args);
		return proxyDB.getFor(value);
	}

	/**
	 * Invokes passed method.
	 * 
	 * @param proxy
	 *            handled proxy object
	 * @param method
	 *            method to invoke
	 * @param args
	 *            argument to pass
	 * @return result
	 * @throws Throwable
	 *             thrown if invocation failed
	 */
	private Object evaluate(Object proxy, Method method, Object[] args)
			throws Throwable {
		if (object.getClass() == method.getDeclaringClass()) {
			if ("toString".equals(method.getName()))
				return proxy.getClass().getName() + "@"
						+ toHexString(identityHashCode(proxy));

			return method.invoke(object, args);
		}
		throw new IllegalStateException(String.valueOf(method));
	}
}
