/**
 * Copyright 2013 Mateusz Kubuszok
 *
 * <p>Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at</p> 
 * 
 * <p>http://www.apache.org/licenses/LICENSE-2.0</p>
 *
 * <p>Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.</p>
 */
package net.jfwu.synchronization;

/**
 * Describes level of synchronization process.
 * 
 * @see net.jfwu.synchronization.SynchronizingProxyHandler
 */
public enum ESynchronizationType {
	/**
	 * No synchronization. Allows concurrent call of method at any time.
	 */
	NONE(false, false),

	/**
	 * Allows concurrent call for all "reading" methods of an object as long as
	 * none is "writing".
	 */
	READ(true, false),

	/**
	 * Allows one "writing" to object - waits for all "reads" to end and locks
	 * object for any other synchronized call until "writing" ends.
	 * 
	 * Same as READ_WRITE but indicates that data is only read.
	 */
	WRITE(false, true),

	/**
	 * Allows one "writing" to object - waits for all "reads" to end and locks
	 * object for any other synchronized call until "writing" ends.
	 * 
	 * Same as READ but indicates that data is both obtained and modified.
	 */
	READ_WRITE(true, true);

	private final boolean read;
	private final boolean write;

	private ESynchronizationType(boolean read, boolean write) {
		this.read = read;
		this.write = write;
	}

	/**
	 * Whether annotated method reads from object.
	 * 
	 * @return true if method reads from object
	 */
	public boolean isRead() {
		return read;
	}

	/**
	 * Whether annotated method modifies object.
	 * 
	 * @return true if method modifies object
	 */
	public boolean isWritten() {
		return write;
	}
}
