/**
 * Copyright 2013 Mateusz Kubuszok
 *
 * <p>Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at</p> 
 * 
 * <p>http://www.apache.org/licenses/LICENSE-2.0</p>
 *
 * <p>Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.</p>
 */
package net.jfwu.synchronization;

import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.util.Map;

import javassist.Modifier;
import javassist.util.proxy.ProxyFactory;

/**
 * 
 * 
 * TODO: adds event handling (start/stop reading/writing)
 */
public final class SynchronizingProxyController {
	private final Map<Object, Object> proxies;

	/**
	 * Creates SynchronizingProxyController.
	 */
	public SynchronizingProxyController() {
		proxies = new HashMap<Object, Object>();
	}

	/**
	 * Makes object synchronized by proxy.
	 * 
	 * @param synchronizedModel
	 *            object to be synchronized
	 */
	public void subscribeForSynchronization(Object synchronizedModel) {
		if (!proxies.containsKey(synchronizedModel))
			proxies.put(synchronizedModel,
					createSynchronizingProxyFor(synchronizedModel));
	}

	/**
	 * Removes object from pool of synchronized objects.
	 * 
	 * @param synchronizedModel
	 *            object to remove from pool
	 */
	public void unsubscribeForSynchronization(Object synchronizedModel) {
		if (proxies.containsKey(synchronizedModel))
			proxies.remove(synchronizedModel);
	}

	/**
	 * Whether object is synchronized.
	 * 
	 * @param synchronizedModel
	 *            object to check
	 * @return true if object is synchronized by controller
	 */
	public boolean isSubscribed(Object synchronizedModel) {
		return proxies.containsKey(synchronizedModel);
	}

	/**
	 * Returns value for an object.
	 * 
	 * <p>
	 * If possible wraps object with a proxy.
	 * </p>
	 * 
	 * @param model
	 *            object to wrap
	 * @return potentially wrapped object
	 */
	public <T> T getFor(T model) {
		if (model != null) {
			if (isSubscribed(model))
				return getProxyFor(model);
			else if (isMockable(model))
				return createPassingProxyFor(model);
		}
		return model;
	}

	/**
	 * Returns SynchronizingProxy for an object.
	 * 
	 * @param synchronizedModel
	 *            object for proxy
	 * @return proxy
	 */
	private <T> T getProxyFor(T synchronizedModel) {
		return (T) proxies.get(synchronizedModel);
	}

	/**
	 * Creates PassingProxy for an object.
	 * 
	 * @param model
	 *            object for proxy
	 * @return proxy
	 */
	private <T> T createPassingProxyFor(T model) {
		ProxyFactory factory = getFactoryFor(model);
		try {
			T t = (T) factory.create(new Class<?>[0], new Object[0],
					new PassingProxyHandler(this, model));
			return t;
		} catch (NoSuchMethodException | IllegalArgumentException
				| InstantiationException | IllegalAccessException
				| InvocationTargetException e) {
			throw new IllegalStateException(e);
		}
	}

	/**
	 * Creates SynchronizingProxy for an object.
	 * 
	 * @param model
	 *            object for proxy
	 * @return proxy
	 */
	private <T> T createSynchronizingProxyFor(T model) {
		ProxyFactory factory = getFactoryFor(model);
		try {
			T t = (T) factory.create(new Class<?>[0], new Object[0],
					new SynchronizingProxyHandler(this, model));
			return t;
		} catch (NoSuchMethodException | IllegalArgumentException
				| InstantiationException | IllegalAccessException
				| InvocationTargetException e) {
			throw new IllegalStateException(e);
		}
	}

	/**
	 * Creates ProxyFactory for an object.
	 * 
	 * @param model
	 *            object for proxy
	 * @return factory
	 */
	private <T> ProxyFactory getFactoryFor(T model) {
		Class<?> clazz = model.getClass();
		if (Modifier.isFinal(clazz.getModifiers()))
			clazz = clazz.getSuperclass();
		ProxyFactory factory = new ProxyFactory();
		factory.setSuperclass(model.getClass());
		return factory;
	}

	/**
	 * Whether proxy can be created for an object.
	 * 
	 * @param model
	 *            object for proxy
	 * @return whether proxy can be created
	 */
	private <T> boolean isMockable(T model) {
		Class<?> clazz = model.getClass();
		return !Modifier.isFinal(clazz.getModifiers())
				&& !Modifier.isPrivate(clazz.getModifiers())
				&& !Modifier.isProtected(clazz.getModifiers());
	}
}
