/**
 * Copyright 2013 Mateusz Kubuszok
 *
 * <p>Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at</p> 
 * 
 * <p>http://www.apache.org/licenses/LICENSE-2.0</p>
 *
 * <p>Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.</p>
 */
package net.jfwu.synchronization;

import static net.jfwu.synchronization.ESynchronizationType.NONE;

import java.lang.reflect.Method;

/**
 * Synchronizes calls to object by configuration passed in annotations.
 * 
 * @see net.jfwu.synchronization.SynchronizedAccess
 * @see net.jfwu.synchronization.SynchronizingProxyHandler
 * @see net.jfwu.synchronization.PassingProxyHandler
 */
class SynchronizingProxyHandler extends AbstractProxyHandler {
	private final Object writingMutex;
	private final Object readingMutex;
	private final Object starvationAvoider;

	private int reads;
	private boolean writing;

	SynchronizingProxyHandler(SynchronizingProxyController proxyDB,
			Object synchronizedModel) {
		super(proxyDB, synchronizedModel);
		this.writingMutex = new Object();
		this.readingMutex = new Object();
		this.starvationAvoider = new Object();
		this.reads = 0;
		this.writing = false;
	}

	@Override
	public Object invoke(Object proxy, Method method, Method superMethod,
			Object[] args) throws Throwable {
		ESynchronizationType accessType = getSynchronizationType(method);

		if (accessType.isWritten()) {
			startWriting();
			waitForEndOfReading();
			try {
				synchronized (writingMutex) {
					return executeCall(proxy, method, args);
				}
			} finally {
				stopWriting();
			}
		} else if (accessType.isRead()) {
			waitForEndOfWritting();
			synchronized (writingMutex) {
				incrementReads();
			}
			try {
				return executeCall(proxy, method, args);
			} finally {
				decrementReads();
			}
		}

		return executeCall(proxy, method, args);
	}

	/**
	 * Decrements number of current reads. When number drops to 0 allows for
	 * writing.
	 */
	private synchronized void decrementReads() {
		reads--;
		if (reads == 0)
			synchronized (readingMutex) {
				readingMutex.notifyAll();
			}
		if (reads < 0)
			throw new IllegalStateException(
					"Number of concurrent reads should never be negative number");
	}

	/**
	 * Increments number of current reads. Till all of them ends no write can
	 * begin.
	 */
	private synchronized void incrementReads() {
		reads++;
	}

	/**
	 * Returns current number of reads.
	 * 
	 * @return reads from object
	 */
	private synchronized int getReads() {
		return reads;
	}

	/**
	 * Wait for end of reading.
	 */
	private void waitForEndOfReading() {
		while (getReads() > 0)
			synchronized (readingMutex) {
				try {
					readingMutex.wait();
				} catch (InterruptedException e) {
				}
			}
	}

	/**
	 * Marks start of writing.
	 */
	private void startWriting() {
		writing = true;
	}

	/**
	 * Marks stop of writing.
	 */
	private void stopWriting() {
		writing = false;
		synchronized (starvationAvoider) {
			starvationAvoider.notifyAll();
		}
	}

	/**
	 * Resumes reading after stop of writing.
	 */
	private void waitForEndOfWritting() {
		if (writing)
			synchronized (starvationAvoider) {
				try {
					starvationAvoider.wait();
				} catch (InterruptedException e) {
				}
			}
	}

	/**
	 * Returns synchronization type for a method.
	 * 
	 * @param method
	 *            potentially synchronized method
	 * @return synchronization level
	 */
	private ESynchronizationType getSynchronizationType(Method method) {
		SynchronizedAccess access = method
				.getAnnotation(SynchronizedAccess.class);
		if (access != null)
			return access.value();
		return NONE;
	}
}
