/**
 * Copyright 2013 Mateusz Kubuszok
 *
 * <p>Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at</p> 
 * 
 * <p>http://www.apache.org/licenses/LICENSE-2.0</p>
 *
 * <p>Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.</p>
 */
package net.jfwu.synchronization;

import static java.lang.System.out;
import static java.lang.Thread.sleep;
import static net.jfwu.synchronization.Action.*;
import static net.jfwu.synchronization.Moment.*;
import static org.fest.assertions.api.Assertions.assertThat;

import org.junit.Before;
import org.junit.Test;

public class TestSynchronization {
	private final String defaultValue = "default";
	private final String newValue = "new";

	private NotSynchronizedClass notSynchronizedClass;
	private SynchronizedClass synchronizedClass;
	private SynchronizingProxyController controller;

	@Before
	public void setUp() {
		controller = new SynchronizingProxyController();
		notSynchronizedClass = new NotSynchronizedClass();
		synchronizedClass = notSynchronizedClass.get();
		controller.subscribeForSynchronization(synchronizedClass);
	}

	@Test
	public void testConcurrentReadsDirectSynchronization() {
		// given
		final SynchronizedClass proxy = controller.getFor(synchronizedClass);
		Thread run1 = directGet(proxy, 0);
		Thread run2 = directGet(proxy, 1);

		// when
		execute(run1, run2);

		// then
		assertThat(synchronizedClass.getHistory()).containsExactly(
				new History(GET, START, defaultValue),
				new History(GET, START, defaultValue),
				new History(GET, END, defaultValue),
				new History(GET, END, defaultValue));
	}

	@Test
	public void testReadThenWriteQueueDirectSynchronization() {
		// given
		final SynchronizedClass proxy = controller.getFor(synchronizedClass);
		Thread run1 = directGet(proxy, 0);
		Thread run2 = directSet(proxy, 1, newValue);

		// when
		execute(run1, run2);

		// then
		assertThat(synchronizedClass.getHistory()).containsExactly(
				new History(GET, START, defaultValue),
				new History(GET, END, defaultValue),
				new History(SET, START, defaultValue),
				new History(SET, END, defaultValue));
		assertThat(synchronizedClass.get()).isEqualTo(newValue);
	}

	@Test
	public void testWriteThenReadQueueDirectSynchronization() {
		// given
		final SynchronizedClass proxy = controller.getFor(synchronizedClass);
		Thread run1 = directSet(proxy, 0, newValue);
		Thread run2 = directGet(proxy, 1);

		// when
		execute(run1, run2);

		// then
		assertThat(synchronizedClass.getHistory()).containsExactly(
				new History(SET, START, defaultValue),
				new History(SET, END, defaultValue),
				new History(GET, START, newValue),
				new History(GET, END, newValue));
		assertThat(synchronizedClass.get()).isEqualTo(newValue);
	}

	@Test
	public void testWriteThenConcurrentReadsDirectSynchronization() {
		// given
		final SynchronizedClass proxy = controller.getFor(synchronizedClass);
		Thread run1 = directSet(proxy, 0, newValue);
		Thread run2 = directGet(proxy, 1);
		Thread run3 = directGet(proxy, 2);

		// when
		execute(run1, run2, run3);

		// then
		assertThat(synchronizedClass.getHistory()).containsExactly(
				new History(SET, START, defaultValue),
				new History(SET, END, defaultValue),
				new History(GET, START, newValue),
				new History(GET, START, newValue),
				new History(GET, END, newValue),
				new History(GET, END, newValue));
		assertThat(synchronizedClass.get()).isEqualTo(newValue);
	}

	@Test
	public void testWriteQueueDirectSynchronization() {
		// given
		final SynchronizedClass proxy = controller.getFor(synchronizedClass);
		Thread run1 = directSet(proxy, 0, newValue);
		Thread run2 = directSet(proxy, 1, defaultValue);

		// when
		execute(run1, run2);

		// then
		assertThat(synchronizedClass.getHistory()).containsExactly(
				new History(SET, START, defaultValue),
				new History(SET, END, defaultValue),
				new History(SET, START, newValue),
				new History(SET, END, newValue));
		assertThat(synchronizedClass.get()).isEqualTo(defaultValue);
	}

	@Test
	public void testNoWritingStarving() {
		// given
		final SynchronizedClass proxy = controller.getFor(synchronizedClass);
		Thread run1 = directGet(proxy, 0);
		Thread run2 = directGet(proxy, 1);
		Thread run3 = directSet(proxy, 2, newValue);
		Thread run4 = directSet(proxy, 3, defaultValue);
		Thread run5 = directGet(proxy, 4);
		Thread run6 = directGet(proxy, 5);

		// when
		execute(run1, run2, run3, run4, run5, run6);

		// then
		if (synchronizedClass.equals(defaultValue))
			assertThat(synchronizedClass.getHistory()).containsExactly(
					new History(GET, START, defaultValue),
					new History(GET, START, defaultValue),
					new History(GET, END, defaultValue),
					new History(GET, END, defaultValue),
					new History(SET, START, defaultValue),
					new History(SET, END, defaultValue),
					new History(SET, START, newValue),
					new History(SET, END, newValue),
					new History(GET, START, defaultValue),
					new History(GET, START, defaultValue),
					new History(GET, END, defaultValue),
					new History(GET, END, defaultValue));
		else
			assertThat(synchronizedClass.getHistory()).containsExactly(
					new History(GET, START, defaultValue),
					new History(GET, START, defaultValue),
					new History(GET, END, defaultValue),
					new History(GET, END, defaultValue),
					new History(SET, START, defaultValue),
					new History(SET, END, defaultValue),
					new History(SET, START, defaultValue),
					new History(SET, END, defaultValue),
					new History(GET, START, newValue),
					new History(GET, START, newValue),
					new History(GET, END, newValue),
					new History(GET, END, newValue));
	}

	@Test
	public void testConcurrentReadsIndirectSynchronization() {
		// given
		final NotSynchronizedClass proxy = controller
				.getFor(notSynchronizedClass);
		Thread run1 = indirectGet(proxy, 0);
		Thread run2 = indirectGet(proxy, 1);

		// when
		execute(run1, run2);

		// then
		assertThat(synchronizedClass.getHistory()).containsExactly(
				new History(GET, START, defaultValue),
				new History(GET, START, defaultValue),
				new History(GET, END, defaultValue),
				new History(GET, END, defaultValue));
	}

	@Test
	public void testReadThenWriteQueueIndirectSynchronization() {
		// given
		final NotSynchronizedClass proxy = controller
				.getFor(notSynchronizedClass);
		Thread run1 = indirectGet(proxy, 0);
		Thread run2 = indirectSet(proxy, 1, newValue);

		// when
		execute(run1, run2);

		// then
		assertThat(synchronizedClass.getHistory()).containsExactly(
				new History(GET, START, defaultValue),
				new History(GET, END, defaultValue),
				new History(SET, START, defaultValue),
				new History(SET, END, defaultValue));
		assertThat(synchronizedClass.get()).isEqualTo(newValue);
	}

	@Test
	public void testWriteThenReadQueueIndirectSynchronization() {
		// given
		final NotSynchronizedClass proxy = controller
				.getFor(notSynchronizedClass);
		Thread run1 = indirectSet(proxy, 0, newValue);
		Thread run2 = indirectGet(proxy, 1);

		// when
		execute(run1, run2);

		// then
		assertThat(synchronizedClass.getHistory()).containsExactly(
				new History(SET, START, defaultValue),
				new History(SET, END, defaultValue),
				new History(GET, START, newValue),
				new History(GET, END, newValue));
		assertThat(synchronizedClass.get()).isEqualTo(newValue);
	}

	@Test
	public void testWriteThenConcurrentReadsIndirectSynchronization() {
		// given
		final NotSynchronizedClass proxy = controller
				.getFor(notSynchronizedClass);
		Thread run1 = indirectSet(proxy, 0, newValue);
		Thread run2 = indirectGet(proxy, 1);
		Thread run3 = indirectGet(proxy, 2);

		// when
		execute(run1, run2, run3);

		// then
		assertThat(synchronizedClass.getHistory()).containsExactly(
				new History(SET, START, defaultValue),
				new History(SET, END, defaultValue),
				new History(GET, START, newValue),
				new History(GET, START, newValue),
				new History(GET, END, newValue),
				new History(GET, END, newValue));
		assertThat(synchronizedClass.get()).isEqualTo(newValue);
	}

	@Test
	public void testQueueWritesIndirectSynchronization() {
		// given
		final NotSynchronizedClass proxy = controller
				.getFor(notSynchronizedClass);
		Thread run1 = indirectSet(proxy, 0, newValue);
		Thread run2 = indirectSet(proxy, 1, defaultValue);

		// when
		execute(run1, run2);

		// then
		assertThat(synchronizedClass.getHistory()).containsExactly(
				new History(SET, START, defaultValue),
				new History(SET, END, defaultValue),
				new History(SET, START, newValue),
				new History(SET, END, newValue));
		assertThat(synchronizedClass.get()).isEqualTo(defaultValue);
	}

	private void execute(Thread... threads) {
		out.println("---Start---");
		for (Thread thread : threads)
			thread.start();
		for (Thread thread : threads)
			try {
				thread.join();
			} catch (InterruptedException e) {
			}
		out.println("---Stop---");
	}

	private void waitAWhile(final int times) {
		try {
			sleep(50 * times);
		} catch (InterruptedException e) {
		}
	}

	private Thread directGet(final SynchronizedClass proxy, final int wait) {
		return new Thread() {
			@Override
			public void run() {
				waitAWhile(wait);
				proxy.get();
			}
		};
	}

	private Thread directSet(final SynchronizedClass proxy, final int wait,
			final String value) {
		return new Thread() {
			@Override
			public void run() {
				waitAWhile(wait);
				proxy.set(value);
			}
		};
	}

	private Thread indirectGet(final NotSynchronizedClass proxy, final int wait) {
		return new Thread() {
			@Override
			public void run() {
				waitAWhile(wait);
				proxy.get().get();
			}
		};
	}

	private Thread indirectSet(final NotSynchronizedClass proxy,
			final int wait, final String value) {
		return new Thread() {
			@Override
			public void run() {
				waitAWhile(wait);
				proxy.get().set(value);
			}
		};
	}
}
